import java.util.Arrays;

/**
 * Class that will present the merge sort algorithm
 *
 * reference: https://www.interviewbit.com/tutorial/merge-sort-algorithm/
 */
public class MergeSort {

    /**
     * Method for merge the sorted result
     *
     * @param inputArray    the given input to merge
     * @param start         the start of the index to merge
     * @param middleOfArray middleOfArray after recurrence split
     * @param end           end of the array for the corresponding array
     */
    private static void merge(int[] inputArray, int start, int middleOfArray, int end) {

        // create a temporary array to store
        int[] tempArray = new int[end - start + 1];

        // crawlers for both intervals and for tempArray
        int index = start, j = middleOfArray + 1, k = 0;

        // traverse both arrays and in each iteration add smaller of both elements in tempArray
        while(index <= middleOfArray && j <= end) {
            if(inputArray[index] <= inputArray[j]) {
                tempArray[k] = inputArray[index];
                k += 1; index += 1;
            } else {
                tempArray[k] = inputArray[j];
                k += 1; j += 1;
            }
        }

        // add elements left in the first interval
        while(index <= middleOfArray) {
            tempArray[k] = inputArray[index];
            k += 1; index += 1;
        }

        // add elements left in the second interval
        while(j <= end) {
            tempArray[k] = inputArray[j];
            k += 1; j += 1;
        }

        // copy tempArray to original interval
        for(index = start; index <= end; index += 1) {
            inputArray[index] = tempArray[index - start];
        }
    }

    /**
     * Merge and sort the array
     *
     * @param inputArray the given input array to sort
     * @param beginning  the beginning of the array
     * @param end        the end index of the array
     */
    private static void mergeAndSort(int[] inputArray, int beginning, int end) {
        if(beginning < end) {
            int midOfArray  = (beginning + end) / 2;
            mergeAndSort(inputArray, beginning, midOfArray);
            mergeAndSort(inputArray, midOfArray + 1, end);
            merge(inputArray, beginning, midOfArray, end);
        }
    }

    public static void main(String[] args) {
        final int[] customArray = new int[]{5, 10, 2, 8, 1, 20, 9, 4};
        mergeAndSort(customArray, 0, customArray.length - 1);
        System.out.println("The sorted array is: " + Arrays.toString(customArray));
    }
}
